import urllib.parse as urlparse
import unicodedata as unicode

from jinja2 import BaseLoader, TemplateNotFound
import lxml.html.html5parser as html5
import lxml.html
import html5lib
import html


def url_to_title(url):
    return url.replace('_', ' ')
def url_to_filename(url):
    return urlparse.quote(unicode.normalize('NFC', url))
def filename_to_url(filename):
    return urlparse.unquote(filename)
def filename_to_title(filename):
    return urlparse.unquote(filename.replace('_', ' '))
def title_to_filename(title):
    return urlparse.quote(unicode.normalize('NFC', title).replace(' ', '_'))

# clean up after Squire, which is generally pretty clean but still needs some help before we can use feed it to pandoc
def sanitize_html(src):
    h = html5.fragment_fromstring(src, create_parent='div')
    brs = h.xpath('//h:br[count(following-sibling::node()) = 0]', namespaces={'h':'http://www.w3.org/1999/xhtml'})
    for br in brs: br.getparent().remove(br)
    # urgh
    walker = html5lib.getTreeWalker("etree")
    stream = walker(h)
    s = html5lib.serializer.HTMLSerializer()
    return ''.join(s.serialize(stream))[5:-6]

class StorageTemplateLoader(BaseLoader):
    def __init__(self, storage):
        self.storage = storage
    
    def get_source(self, environment, template):
        latest = self.storage.latest()
        templates = latest.dir('templates')
        
        try:
            source = templates.get(template).decode('utf-8')
        except FileNotFoundError:
            raise TemplateNotFound(template)
        
        return source, None, lambda: latest.revision == self.storage.latest().revision



import mimetypes
import os
import os.path
from urllib.parse import urlparse, urljoin
import yaml
import random

import bcrypt
from jinja2 import Environment
import pypandoc

# Save PEP 3122!
if "." in __name__:
    from .storage import Storage, Signature
    from .util import url_to_title, url_to_filename, title_to_filename, sanitize_html, StorageTemplateLoader
    from .www import Application, Request, Response, JSONResponse
    from .search import LinksDatabase, SearchDatabase
else:
    from storage import Storage, Signature
    from util import url_to_title, url_to_filename, title_to_filename, sanitize_html, StorageTemplateLoader
    from www import Application, Request, Response, JSONResponse
    from search import LinksDatabase, SearchDatabase


class Ikwi(Application):
    image_extensions = ['.jpg', '.png', '.svg', '.gif']
    version = '0.1'
    
    def __init__(self, repo_path):
        self.storage = Storage(repo_path)
        
        self.base_url = ''
        self.base_path = ''
        self.config = {}
        self.config_revision = None
        self.jinja_env = Environment(
            loader=StorageTemplateLoader(self.storage),
            autoescape=True
        )
        self.jinja_env.globals = {
            'site_url': self.site_url
        }
        
        self.links = LinksDatabase(self)
        self.search = SearchDatabase(self)

    def before_request(self, request):
        self.latest = self.storage.latest()
        if self.latest.revision != self.config_revision:
            self.config = yaml.load(self.latest.get('site.yaml').decode('utf-8'))
            if 'base_url' in self.config:
                self.base_url = self.config['base_url']
                self.base_path = urlparse(self.base_url).path.rstrip('/')
            else:
                self.base_url = '/'
        request.path = request.path[len(self.base_path):]

    def site_url(self, path=''):
        return urljoin(self.base_url, path)

    def is_internal_link(self, path):
        if path.startswith(self.base_url):
            internal_path = path[len(self.base_url):].strip('/')
        elif path.startswith('/' + self.base_path):
            internal_path = path[len(self.base_path)+1:].strip('/')
        elif not path.startswith('/'):
            internal_path = path
        else:
            return False
        
        parts = internal_path.split('/')
        if len(parts) == 0: return True # link to homepage
        if parts[0] in {'files', 'images', 'site'}:
            return False
        else:
            return True

    def render_template(self, template_name, **context):
        t = self.jinja_env.get_template(template_name)
        vars = self.config.copy()
        vars.update(context)
        
        return Response(t.render(vars), mimetype='text/html')

    def dispatch_request(self, request):
        base, *path = request.path.strip('/').split('/')
        
        if base == 'files':
            self.require_method(request, ['GET'])
            return self.serve_file(path, request)
        elif base == 'images':
            self.require_method(request, ['GET'])
            if len(path) == 0: return self.not_found()
            
            if request.query_verb == 'old' and 'rev' in request.args:
                old = self.storage.at_revision(request.args['rev'])
                return self.serve_image(path[0], old, request)
            else:
                return self.serve_image(path[0], self.latest, request)
        elif base == 'site':
            self.require_method(request, ['GET'])
            if path == ['edit.js']:
                js_dir = os.path.join(os.path.dirname(__file__), 'js')
                js_files = ['squire.js', 'jquery.js', 'underscore.js', 'editor.js']
                def js_gen():
                    for js_filename in js_files:
                        with open(os.path.join(js_dir, js_filename), 'rb') as file:
                            yield from file
                            yield b'\n'
            
                response = Response(js_gen(), mimetype='application/javascript')
                #response.set_etag(Ikwi.version)
                #response.make_conditional(request)
                return response
            if path == ['search']:
                results = self.search.search(request.args['q'])
                return JSONResponse({'query': request.args['q'], 'results': results})
            else:
                return self.not_found()
        else:
            self.require_method(request, ['GET', 'POST'])
            if len(path) > 0: return self.not_found()
            
            url_page_name = (base or 'Homepage')
            if request.method == 'GET':
                if request.query_verb == 'old':
                    return self.show_page(url_page_name, self.storage.at_revision(request.args['rev']))
                elif request.query_verb == 'edit':
                    self.must_login(request)
                    return self.edit_page(url_page_name)
                elif request.query_verb == 'inlinks':
                    return self.show_inlinks(url_page_name)
                elif request.query_verb in {None, 'no-redirect'}:
                    return self.show_page(url_page_name, self.latest)
                else:
                    return self.not_found()
            elif request.method == 'POST':
                self.must_login(request)
                return self.save_page(url_page_name, request)

    def to_html(self, source):
        return pypandoc.convert(source, 'html', format=self.config['page_format'])
    def to_source(self, html):
        return pypandoc.convert(html, self.config['page_format'], format='html')

    def get_page(self, url_page_name, revision):
        pages = revision.dir('pages')
        filename = url_to_filename(url_page_name)
        if filename in pages:
            return pages.get(filename)
        else:
            return None

    def header_image(self, page_filename, revision):
        if revision.revision != self.latest.revision:
            old_string = '?old&rev=%s' % revision.revision
        else:
            old_string = ''
        
        images = revision.dir('images')
        for extension in Ikwi.image_extensions:
            image_filename = page_filename + extension
            if image_filename in images:
                return self.site_url('images/' + image_filename + old_string)

    def show_page(self, url_page_name, revision):
        page_title = url_to_title(url_page_name)
        page_source = self.get_page(url_page_name, revision)
        
        if not page_source:
            return self.not_found(creatable=True)
        
        page_content = self.to_html(page_source)
        header_image = self.header_image(url_to_filename(url_page_name), revision)
        
        response = self.render_template('page.html', page_title=page_title, page_content=page_content, header_image=header_image)
        # todo: make response cacheable
        return response
    
    def edit_page(self, url_page_name):
        page_title = url_to_title(url_page_name)
        page_source = self.get_page(url_page_name, self.latest)
        
        if not page_source:
            page_content = '<p></p>'
        else:
            page_content = self.to_html(page_source)
        
        header_image = self.header_image(url_to_filename(url_page_name), self.latest)
        
        return self.render_template('edit.html', page_title=page_title, page_content=page_content, header_image=header_image, revision_id=self.latest.revision)
    
    def save_page(self, filename, request):
        title = request.form['title']
        html = sanitize_html(request.form['content'])
        filename = title_to_filename(title)
        
        cursor = self.storage.cursor(request.form['revision'])
        cursor.add('pages/' + filename, self.to_source(html).encode('utf-8'))
        
        if 'headerimage' in request.files:
            header_image = request.files['headerimage'].read()
            extension = mimetypes.guess_extension(request.files['headerimage'].mimetype)
            if extension.startswith('.jpe'): extension = '.jpg' # wtf Python?!
            
            if extension in Ikwi.image_extensions:
                image_filename = filename + extension
                for other_extension in Ikwi.image_extensions:
                    cursor.delete('images/' + filename + other_extension)
                
                cursor.add('images/' + image_filename, header_image)
            
        cursor.save('%s: %s' % (title, request.form.get('change_message', '')), Signature(self.config['editors'][request.authorization.username]['name'], self.config['editors'][request.authorization.username]['email']))
        status = cursor.update('HEAD')
        
        if status.conflict:
            return JSONResponse({
                'status': 'conflict',
                'source': status.source_revision,
                'target': status.target_revision,
            }, 409)
        
        return JSONResponse({'status': 'ok', 'revision': status.revision})

    def show_inlinks(self, url_page_name):
        page_title = url_to_title(url_page_name)
        filename = url_to_filename(url_page_name)
        inlinks = self.links.inlinks(filename)
        return self.render_template('inlinks.html', inlinks=inlinks, page_title=page_title, page_url=url_page_name)

    def serve_file(self, path, request):
        path = path[0]
        dir = self.latest.dir('files')
        if path not in dir:
            return self.not_found()
        
        type, encoding = mimetypes.guess_type(path)
        
        # prevent the blob from being decoded unless actually needed
        def yield_get(): yield dir.get(path)
        response = Response(yield_get(), mimetype=type, direct_passthrough=True)
        response.set_etag(dir.get_id(path))
        response.make_conditional(request)
        return response

    def serve_image(self, path, revision, request):
        path = url_to_filename(path)
        dir = self.latest.dir('images')
        if path not in dir:
            return self.not_found()
        type, encoding = mimetypes.guess_type(path)
        
        # prevent the blob from being decoded unless actually needed
        def yield_get(): yield dir.get(path)
        response = Response(yield_get(), mimetype=type, direct_passthrough=True)
        response.set_etag(dir.get_id(path))
        response.make_conditional(request)
        return response

    def must_login(self, request):
        if not request.authorization:
            raise PermissionError
        
        username = request.authorization.username
        try_password = request.authorization.password.encode('utf-8')
        if username not in self.config['editors']:
            raise PermissionError
        
        real_password = self.config['editors'][username]['password'].encode('utf-8')
        if bcrypt.hashpw(try_password, real_password) != real_password:
            raise PermissionError
    
    def unauthorized(self):
        response = self.render_template('unauthorized.html')
        response.headers.extend({
            'WWW-Authenticate': 'Basic realm="%s"' % self.config['site_title']
        })
        return Response(
            response.response,
            401,
            response.headers
        )
    
    def not_found(self, creatable=False):
        response = self.render_template('not_found.html', creatable=creatable)
        return Response(
            response.response,
            404,
            response.headers
        )



import os
import os.path
import sqlite3
from urllib.parse import urlparse

import lxml.html.html5parser as html5
import whoosh.fields
import whoosh.index
import whoosh.qparser

if "." in __name__:
    from .database import Database
    from .util import url_to_filename, filename_to_title, filename_to_url
else:
    from database import Database
    from util import url_to_filename, filename_to_title, filename_to_url


ns = {'h':'http://www.w3.org/1999/xhtml'}

class LinksDatabase(Database):
    database_name = 'links.sqlite3'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conn = sqlite3.connect(self.path)

    def do_create(self):
        c = self.conn.cursor()
        c.execute("CREATE TABLE links (id INTEGER PRIMARY KEY, source TEXT NOT NULL, target TEXT NOT NULL)")
        c.execute("CREATE INDEX outlinks ON links (source)")
        c.execute("CREATE INDEX inlinks ON links (target)")
        
        c.execute("CREATE TABLE redirects (source TEXT PRIMARY KEY NOT NULL, target TEXT NOT NULL)")
        c.execute("CREATE INDEX aliases ON redirects (target)")
        
        c.execute("CREATE TABLE pageranks (page TEXT PRIMARY KEY NOT NULL, rank REAL NOT NULL)")
        self.conn.commit()
    
    def do_update(self, differences):
        c = self.conn.cursor()
        
        def delete_page(page):
            c.execute('DELETE FROM links INDEXED BY outlinks WHERE source = ?', (page,))
            c.execute('DELETE FROM redirects WHERE source = ?', (page,))
        def index_content(page, content):
            content = content.decode('utf-8')
            if content.startswith('=> '):
                c.execute('INSERT INTO redirects (source, target) VALUES (?, ?)', (page, content[3:].strip()))
                return
            
            src = html5.fragment_fromstring(self.site.to_html(content), create_parent='div')
            links = src.xpath('//h:a[@href]', namespaces=ns)
            page_links = []
            
            for link in links:
                dest = link.attrib['href']
                destinfo = urlparse(dest)
                if self.site.is_internal_link(dest):
                    if dest.startswith('http:') or dest.startswith('https:'):
                        dest_name = url_to_filename(destinfo.path.strip('/'))
                    elif dest.startswith('/' + self.site.base_path):
                        dest_name = url_to_filename(destinfo.path[len(self.site.base_path)+2:])
                    else:
                        dest_name = url_to_filename(destinfo.path.strip('/'))
                    
                    if dest_name == '': dest_name = 'Homepage'
                    page_links.append((page, dest_name))
            
            c.executemany('INSERT INTO links (source, target) VALUES (?, ?)', page_links)
        
        try:
            for page, difference in differences.items():
                op, content = difference
                if op == 'created':
                    index_content(page, content)
                elif op == 'updated':
                    delete_page(page)
                    index_content(page, content)
                elif op == 'deleted':
                    delete_page(page)
        except:
            self.conn.rollback()
            raise
        
        self.conn.commit()

    def inlinks(self, filename):
        self.update()
        c = self.conn.cursor()
        c.execute('SELECT source FROM links WHERE target = ?', (filename,))
        for result in c:
            yield Link(filename_to_url(result[0]), filename_to_title(result[0]))

class Link:
    def __init__(self, url, title):
        self.url = url
        self.title = title

class SearchDatabase(Database):
    database_name = 'search.whoosh'
    schema = whoosh.fields.Schema(
        filename=whoosh.fields.ID(stored=True, unique=True),
        url=whoosh.fields.STORED,
        title=whoosh.fields.TEXT(field_boost=2.0, stored=True),
        content=whoosh.fields.TEXT,
        redirect_to=whoosh.fields.STORED
    )
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if whoosh.index.exists_in(self.path):
            self.index = whoosh.index.open_dir(self.path)
        else:
            if not os.path.isdir(self.path): os.mkdir(self.path)
            self.index = None
    
    def do_create(self):
        self.index = whoosh.index.create_in(self.path, SearchDatabase.schema)
    
    def do_update(self, differences):
        self.index = self.index.refresh()
        with self.index.writer() as writer:
            for page, difference in differences.items():
                op, content = difference
                
                if content:
                    content = content.decode('utf-8')
                    if content.startswith('=> '):
                        redirect_to = content[3:].strip()
                        doc = {
                            'filename': page,
                            'url': filename_to_url(page),
                            'title': filename_to_title(page),
                            'content': '',
                            'redirect_to': redirect_to
                        }
                    else:
                        src = html5.fragment_fromstring(self.site.to_html(content), create_parent='div')
                        content = ' '.join(src.xpath('//text()'))
                        doc = {
                            'filename': page,
                            'url': filename_to_url(page),
                            'title': filename_to_title(page),
                            'content': content,
                            'redirect_to': None
                        }
                
                if op == 'created':
                    writer.add_document(**doc)
                elif op == 'updated':
                    writer.update_document(**doc)
                elif op == 'deleted':
                    writer.delete_by_term('filename', page)

    def search(self, query):
        self.update()
        self.index = self.index.refresh()
        parser = whoosh.qparser.MultifieldParser(['title', 'content'], SearchDatabase.schema)
        parsed_query = parser.parse(query)
        with self.index.searcher() as searcher:
            results = searcher.search(parsed_query, limit=30)
            return list(dict(result) for result in results)
        



# Copyright (c) 2015 Tony Garnock-Jones <tonyg@leastfixedpoint.com>
# Copyright (c) 2015 Sean B. Palmer <inamidst.com>
# Licensed under the 2-Clause Simplified BSD License

def longest_common_subsequence(xs, ys):
    # Uses the Ukkonen-Myers algorithm
    t = len(xs) + len(ys)
    front = [0] * (2 * t + 1)
    candidates = [None] * (2 * t + 1)
    for d in range(t + 1):
        for k in range(-d, d+1, 2):
            if k == -d or (k != d and front[t + k - 1] < front[t + k + 1]):
                index = t + k + 1
                x = front[index]
            else:
                index = t + k - 1
                x = front[index] + 1
            y = x - k
            chain = candidates[index]
            while x < len(xs) and y < len(ys) and xs[x] == ys[y]:
                chain = ((x, y), chain)
                x += 1
                y += 1
            if x >= len(xs) and y >= len(ys):
                result = []
                while chain:
                    result.append(chain[0])
                    chain = chain[1]
                result.reverse()
                return result
            front[t + k] = x
            candidates[t + k] = chain

def diff(xs, ys):
    i = -1
    j = -1
    matches = longest_common_subsequence(xs, ys)
    matches.append((len(xs), len(ys)))
    result = []
    for (mi, mj) in matches:
        if mi - i > 1 or mj - j > 1:
            result.append((i + 1, xs[i + 1:mi], j + 1, ys[j + 1:mj]))
        i = mi
        j = mj
    return result

def patch(a, p, reverse=False):
    pat = {}
    for (ai, astr, bi, bstr) in p:
        if reverse is True:
            ai, astr, bi, bstr = bi, bstr, ai, astr
        if ai in pat:
           raise ValueError("Invalid patch")
        pat[ai] = (len(astr), bstr)

    i = 0
    parts = []
    while i < (len(a) + 1):
        skip, replacement = pat.get(i, (1, a[i:i+1]))
        parts.append(replacement)
        i += skip
        if skip == 0:
            del pat[i]
    return "".join(parts)

def example():
    a = "The red brown fox jumped over the rolling log"
    b = "The brown spotted fox leaped over the rolling log"
    print(patch(b, diff(a, b), True))
    print(patch(a, diff(a, b)))
    print(diff(a, b))

if __name__ == "__main__":
    example()



"""
storage -- tidemark git in a way that's useful for building web-apps
"""
# ᚾᚪᚻᛏ᛫ᛁᛋ᛫ᚦᚱᚫᛞᛋᛁᚳᚩᚱ
# todo: straighten out the terminology used here ('store' and 'storage'; 'commit_id', 'commit', and 'revision')
# merge conflict resolution API probably needs a little tweaking

from collections import deque
import os
import os.path
import time
import warnings

import pygit2


def commit_id(id):
    if isinstance(id, pygit2.Oid):
        return str(id)
    elif isinstance(id, str):
        return id
    elif isinstance(id, bytes):
        return ''.join('%02x' % byte for byte in id)
    else:
        raise TypeError('commit_id() argument must be Oid, str, or bytes')

class NoConflictError(Exception): pass
class Storage:
    def __init__(self, repo_path):
        self.repo = pygit2.Repository(repo_path)
    
    def cursor(self, base_commit):
        return Cursor(self, commit_id(base_commit))
    
    def latest(self):
        head_commit = self.repo[self.repo.head.resolve().target]
        return StorageRevision(self, head_commit.id, head_commit.tree)
    
    def merge_conflict(self, source_revision, target_revision):
        merge = self.repo.merge_commits(target_revision, source_revision)
        if not merge.conflicts:
            raise NoConflictError('an attempt was made to resolve a conflict when merging from %r into %r, but no conflict exists between those revisions' % (source_revision, target_revision))
        
        return MergeConflict(self, source_revision, target_revision, merge.conflicts)
    
    def ref_lock(self, ref, *, spin_tries=0, spin_wait=0.2):
        lock = None
        while spin_tries >= 0:
            try:
                lock = ReferenceLock(self.repo, ref)
                break
            except ReferenceLockedError:
                if spin_tries == 0:
                    raise
                else:
                    spin_tries -= 1
                    time.sleep(spin_wait)
        
        return lock

class StorageRevision:
    def __init__(self, storage, revision, tree, root_tree=None):
        self.storage = storage
        self.revision = commit_id(revision)
        self.tree = tree
        if not root_tree:
            self.root_tree = tree
        else:
            self.root_tree = root_tree
    
    def __contains__(self, filename):
        return (filename in self.tree) and (self.tree[filename].filemode == pygit2.GIT_FILEMODE_BLOB)
    
    def get_id(self, filename):
        tree_entry = self.tree[filename]
        return str(tree_entry.id)
    
    def get(self, filename):
        if filename not in self.tree: return None
        tree_entry = self.tree[filename]
        if tree_entry.filemode != pygit2.GIT_FILEMODE_BLOB: return None
        blob = self.storage.repo[tree_entry.id]
        return blob.data
    
    def dir(self, dirname):
        if dirname not in self.tree: return EmptyStorageRevision()
        tree_entry = self.tree[dirname]
        if tree_entry.filemode != pygit2.GIT_FILEMODE_TREE: return EmptyStorageRevision()
        tree = self.storage.repo[tree_entry.id]
        return StorageRevision(self.storage, self.revision, tree, self.root_tree)

# this needs a better API
class EmptyStorageRevision:
    def get(self, filename): return None

class InvalidOperationError(Exception): pass
class Cursor:
    def __init__(self, storage, base_commit, *, original_base_commit=None):
        self.storage = storage
        self.repo = self.storage.repo
        self.base_commit_id = commit_id(base_commit)
        
        if original_base_commit:
            self.original_base_commit_id = commit_id(original_base_commit)
        else:
            self.original_base_commit_id = commit_id(base_commit)
        
        self.root_tree = self.repo[base_commit].tree
    
    def add(self, path, contents):
        blob_id = self.repo.create_blob(contents)
        idx = pygit2.Index()
        idx.read_tree(self.root_tree)
        
        invalid = False
        try:
            idx.add(pygit2.IndexEntry(path, blob_id, pygit2.GIT_FILEMODE_BLOB))
        except pygit2.GitError as err:
            if 'appears as both a file and a directory' in err.args[0]:
                invalid = True
        
        if invalid:
            raise InvalidOperationError('refusing to replace another kind of object with a tree')

        tree_id = idx.write_tree(self.repo)
        self.root_tree = self.repo[tree_id]

    def delete(self, path):
        idx = pygit2.Index()
        idx.read_tree(self.root_tree)
        
        try:
            idx.remove(path)
        except OSError:
            return False
        
        tree_id = idx.write_tree(self.repo)
        self.root_tree = self.repo[tree_id]
        return True
    
    def save(self, message, author, committer=None):
        if committer == None: committer = author
        new_commit_id = self.repo.create_commit(
            None, # reference
            author,
            committer,
            message,
            self.root_tree.id, # tree
            [self.base_commit_id] # parents
        )
        self.base_commit_id = commit_id(new_commit_id)
    
    def update(self, ref='HEAD', *, merger=None, spin_tries=10, spin_wait=0.2):
        if not merger:
            committer = self.repo[self.base_commit_id].committer
            merger = Signature(committer.name, committer.email)
        
        with self.storage.ref_lock(ref, spin_tries=spin_tries, spin_wait=0.2) as lock:
            if commit_id(self.repo.lookup_reference(lock.ref_name).target) == self.original_base_commit_id:
                # no other changes meanwhile
                lock.set_target(self.base_commit_id)
                return NoMergeWasNeeded(self.repo[self.base_commit_id])
            else:
                merge = self.repo.merge_commits(self.repo.lookup_reference(lock.ref_name).target, self.base_commit_id)
                if merge.conflicts:
                    return MergeConflict(self.storage, self.base_commit_id, self.repo.lookup_reference(lock.ref_name).target, merge.conflicts)
                else:
                    merge_tree = merge.write_tree(self.repo)
                    merge_commit_id = self.repo.create_commit(
                        None, # reference
                        merger, # author
                        merger, # committer
                        ('Merge commit %r into %r' % (self.base_commit_id, lock.ref_name)), # message
                        merge_tree, # tree
                        [self.repo.lookup_reference(lock.ref_name).target, self.base_commit_id] # parents
                    )
                    lock.set_target(merge_commit_id)
                    return AutoMerged(self.repo[merge_commit_id])

class UpdateResult:
    def __init__(self, commit):
        self.commit = commit
        self.revision = commit_id(commit.id)
    
    @property
    def merged(self): return len(self.commit.parents) > 1
    
    @property
    def conflict(self): return False

class NoMergeWasNeeded(UpdateResult): pass
class AutoMerged(UpdateResult): pass

class MergeConflictUnresolved(Exception): pass
class MergeConflict(UpdateResult):
    def __init__(self, store, source_revision, target_revision, conflicts):
        self.store = store
        self.source_revision = commit_id(source_revision)
        self.target_revision = commit_id(target_revision)
        
        self.conflicts = {}
        for original, target_version, source_version in conflicts:
            def version_content(version):
                if version is None: return None
                return self.store.repo[version.hex].data
            
            path = [v for v in (original, target_version, source_version) if v][0].path
            self.conflicts[path] = ConflictedFile(
                original=version_content(original),
                source_version=version_content(source_version),
                target_version=version_content(target_version)
            )

    @property
    def conflict(self): return True

    def resolve(self, resolutions, merger=None):
        repo = self.store.repo
        if not merger:
            committer = repo[self.source_revision].committer
            merger = Signature(committer.name, committer.email)
        
        merge = self.store.repo.merge_commits(self.source_revision, self.target_revision)
        
        for path, resolution in resolutions.items():
            blob_id = repo.create_blob(resolution)
            merge.add(pygit2.IndexEntry(path, blob_id, pygit2.GIT_FILEMODE_BLOB))
            del merge.conflicts[path]
        
        if merge.conflicts:
            raise MergeConflictUnresolved('not all the conflicts between %r and %r were resolved' % (source_revision, target_revision))
        
        merge_tree = merge.write_tree(repo)
        merge_commit_id = repo.create_commit(
            None, # reference
            merger,
            merger,
            ('Merge commit %r with %r' % (self.source_revision, self.target_revision)),
            merge_tree,
            [self.target_revision, self.source_revision]
        )
        
        return Cursor(self.store, merge_commit_id, original_base_commit=self.target_revision)

class ConflictedFile:
    __all__ = ('original', 'source_version', 'target_version')
    def __init__(self, original, source_version, target_version):
        self.original = original
        self.source_version = source_version
        self.target_version = target_version
    
    def __eq__(self, other):
        if not isinstance(other, ConflictedFile): return False
        return (self.original == other.original) and (self.source_version == other.source_version) and (self.target_version == other.target_version)

# we might need to overload more functionality onto Signature one day, but for now this will do:
Signature = pygit2.Signature

# when pygit2 implements transactions we can get rid of this
class ReferenceLockedError(Exception): pass
class UnsavedReferenceLockWarning(Warning): pass
class ReferenceLock:
    def __init__(self, repo, ref_name):
        self.repo = repo
        # canonicalize the name
        ref = self.repo.lookup_reference(ref_name).resolve()
        self.ref_name = ref.name
        self.original_target = commit_id(ref.target)
        
        self.lock_file_path = os.path.join(self.repo.path, self.ref_name + '.lock')
        try:
            self.lock_file = open(self.lock_file_path, 'x', encoding='us-ascii')
        except FileExistsError:
            raise ReferenceLockedError('reference %r on repo %r is already locked' % (self.ref_name, self.repo.path))
        
        self.set_target(ref.target)
    
    def set_target(self, target):
        target = commit_id(target)
        self.lock_file.seek(0)
        print(target, file=self.lock_file)
        # for safety:
        self.lock_file.flush()
        self.lock_file.truncate(self.lock_file.tell())
    
    def revert(self): return self.set_target(self.original_target)
    
    def save(self):
        self.lock_file.close()
        os.rename(self.lock_file_path, os.path.join(self.repo.path, self.ref_name))
    
    def __enter__(self): return self
    def __exit__(self, *exc_info): self.save()
    
    def __del__(self):
        if not self.lock_file.closed:
            self.lock_file.close()
            os.remove(self.lock_file_path)
            warnings.warn('a reference lock for %r on repository %r was deleted without being saved' % (self.ref_name, self.repo.path), UnsavedReferenceLockWarning)



"""
www -- minimal web framework, a customized version of Werkzeug
"""
import json

from werkzeug.wrappers import Request as BaseRequest, Response
from werkzeug.serving import run_simple
from werkzeug.datastructures import ImmutableOrderedMultiDict


class Request(BaseRequest):
    parameter_storage_class = ImmutableOrderedMultiDict
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        args_iterator = iter(self.args.items())
        try:
            key, value = next(args_iterator)
            if value == '':
                self.query_verb = key
                self.args = ImmutableOrderedMultiDict(args_iterator)
            else:
                self.query_verb = None
        except StopIteration:
            self.query_verb = None

# class Response(BaseResponse): pass

def JSONResponse(obj, code=200, headers={}):
    return Response(
        json.dumps(obj),
        code,
        headers.update({'Content-Type': 'application/json'})
    )

class MethodNotAllowed(Exception): pass
class Application:
    def wsgi_app(self, environ, start_response):
        request = Request(environ)
        self.before_request(request)
        
        try:
            response = self.dispatch_request(request)
        except FileNotFoundError:
            response = self.not_found()
        except PermissionError:
            response = self.unauthorized()
        except MethodNotAllowed:
            response = Response('Method %s is not allowed on this resource.' % (request.method), 405)
        
        return response(environ, start_response)

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def run(self):
        run_simple('127.0.0.1', 3000, self, use_debugger=True)
    
    def require_method(self, request, methods):
        if request.method not in methods:
            raise MethodNotAllowed



# this is not an implementation of PageRank, merely a description of PageRank
import sys

def pagerank(graph, damping=0.85, epsilon=1.0e-8):
    inlink_map = {}
    outlink_counts = {}
    
    def new_node(node):
        if node not in inlink_map: inlink_map[node] = set()
        if node not in outlink_counts: outlink_counts[node] = 0
    
    for tail_node, head_node in graph:
        new_node(tail_node)
        new_node(head_node)
        if tail_node == head_node: continue
        
        if tail_node not in inlink_map[head_node]:
            inlink_map[head_node].add(tail_node)
            outlink_counts[tail_node] += 1
    
    all_nodes = set(inlink_map.keys())
    for node, outlink_count in outlink_counts.items():
        if outlink_count == 0:
            outlink_counts[node] = len(all_nodes)
            for l_node in all_nodes: inlink_map[l_node].add(node)
    
    initial_value = 1 / len(all_nodes)
    ranks = {}
    for node in inlink_map.keys(): ranks[node] = initial_value
    
    new_ranks = {}
    delta = 1.0
    n_iterations = 0
    while delta > epsilon:
        for node, inlinks in inlink_map.items():
            new_ranks[node] = ((1 - damping) / len(all_nodes)) + (damping * sum(ranks[inlink] / outlink_counts[inlink] for inlink in inlinks))
        delta = abs(sum(new_ranks.values()) - sum(ranks.values()))
        ranks = new_ranks
        n_iterations += 1
    
    return ranks, n_iterations

if __name__ == '__main__':
    graph = (line.strip().split(None, 1) for line in open(sys.argv[1], 'r'))
    ranks, n_iterations = pagerank(graph)
    
    print('%d iterations' % n_iterations, file=sys.stderr)
    ordered = sorted(ranks.items(), key=lambda node: node[1], reverse=True)
    for node, rank in ordered:
        print(node, rank)



#!/usr/bin/env python

from distutils.core import setup

setup(
    name='ikwi',
    version='0.1.4',
    description='A personal wikiware.',
    author='David P. Kendal',
    author_email='pypi@dpk.io',
    url='https://github.com/dpk/ikwi',
    packages=['ikwi'],
    package_dir={'ikwi': '.'},
    package_data={'ikwi': ['js/*.js']},
    scripts=['ikwi'],
    install_requires=[
        'bcrypt>=1.1.1',
        'html5lib==0.999',
        'Jinja2>=2.7.3',
        'lxml>=3.4.3',
        'pygit2==0.22.0',
        'pypandoc==0.9.7',
        'PyYAML==3.11',
        'Werkzeug>=0.10.4',
        'whoosh>=2.7.0'
    ]
)



# TODO: implement this only in terms of storage, not using Git directly
import os
import os.path
import time

# common functionality for the search and links databases
class Database:
    def __init__(self, site):
        self.site = site

    @property
    def path(self):
        return os.path.join(self.site.storage.repo.path, type(self).database_name) # come at me, Demeter!
    
    @property
    def outdated(self):
        try:
            with open(self.path + '.head', 'r', encoding='us-ascii') as f:
                db_version = f.read().strip()
                site_version = self.site.latest.revision
                return db_version != site_version
        except FileNotFoundError:
            return True
    
    @property
    def current_version(self):
        with open(self.path + '.head', 'r', encoding='us-ascii') as f:
            return f.read().strip()
    
    def update(self, tries=20):
        if not self.outdated: return
        try:
            lock_file = None
            try:
                lock_file = open(self.path + '.head.lock', 'x', encoding='us-ascii') 
                repo = self.site.storage.repo
                
                try:
                    db_version = self.current_version
                    
                    print(db_version, file=lock_file)
                    lock_file.flush()
                    lock_file.seek(0)
                
                    old_tree = repo[repo[db_version].tree['pages'].id]
                    new_tree = repo[repo[self.site.latest.revision].tree['pages'].id]
                
                    differences = {}
                    for page in new_tree:
                        if page.name not in old_tree:
                            differences[page.name] = ('created', repo[page.id].data)
                        elif old_tree[page.name].id != new_tree[page.name].id:
                            differences[page.name] = ('updated', repo[page.id].data)
                
                    for page in old_tree:
                        if page.name not in new_tree:
                            differences[page.name] = ('deleted', None)
                except FileNotFoundError:
                    self.do_create()
                    tree = repo[repo[self.site.latest.revision].tree['pages'].id]
                    differences = {}
                    for page in tree:
                        differences[page.name] = ('created', repo[page.id].data)
                
                self.do_update(differences)
                
                print(self.site.latest.revision, file=lock_file)
                lock_file.flush()
            finally:
                if lock_file and not lock_file.closed:
                    lock_file.close()
                os.rename(self.path + '.head.lock', self.path + '.head')
        except FileExistsError:
            if tries > 0:
                time.sleep(0.1)
                self.update(tries=tries - 1)
            else:
                raise
